@@ -1,6 +1,6 @@
 /* -*- c-file-style: "ruby"; indent-tabs-mode: nil -*- */
 /*
- *  Copyright (C) 2002-2018  Ruby-GNOME2 Project Team
+ *  Copyright (C) 2002-2021  Ruby-GNOME Project Team
  *  Copyright (C) 2002,2003  Masahiro Sakai
  *
  *  This library is free software; you can redistribute it and/or
@@ -31,9 +31,10 @@ static VALUE rb_cMutex;
 #endif
 static VALUE lookup_class_mutex;
 
+static ID id_gtype;
+static ID id_lock;
 static ID id_new;
 static ID id_superclass;
-static ID id_lock;
 static ID id_unlock;
 static GHashTable *gtype_to_cinfo;
 static VALUE klass_to_cinfo;
@@ -99,6 +100,29 @@ rbgobj_class_info_fill_name(RGObjClassInfo *cinfo)
     cinfo->data_type->wrap_struct_name = cinfo->name;
 }
 
+static rb_data_type_t *
+rbgobj_class_info_create_data_type(VALUE klass)
+{
+    rb_data_type_t *data_type;
+
+    data_type = RB_ZALLOC(rb_data_type_t);
+    data_type->function.dmark = cinfo_mark;
+    data_type->function.dfree = cinfo_free;
+    if (RB_TYPE_P(klass, RUBY_T_CLASS) && klass != rb_cObject) {
+        VALUE p = RCLASS_SUPER(klass);
+        while (p != rb_cObject) {
+            if (RB_TYPE_P(p, RUBY_T_DATA) && RTYPEDDATA_P(p)) {
+                data_type->parent = RTYPEDDATA_TYPE(p);
+                break;
+            }
+            p = RCLASS_SUPER(p);
+        }
+    }
+    data_type->flags = RUBY_TYPED_FREE_IMMEDIATELY;
+
+    return data_type;
+}
+
 static RGObjClassInfo *
 rbgobj_class_info_register_without_lock(GType gtype, VALUE klass)
 {
@@ -109,22 +133,8 @@ rbgobj_class_info_register_without_lock(GType gtype, VALUE klass)
     void *gclass = NULL;
     VALUE c;
 
-    data_type = RB_ZALLOC(rb_data_type_t);
-    data_type->function.dmark = cinfo_mark;
-    data_type->function.dfree = cinfo_free;
-    if (RB_TYPE_P(klass, RUBY_T_CLASS)) {
-        VALUE p = RCLASS_SUPER(klass);
-        while (p != rb_cObject) {
-            if (RTYPEDDATA_P(p)) {
-                data_type->parent = RTYPEDDATA_TYPE(p);
-                break;
-            }
-            p = RCLASS_SUPER(p);
-        }
-    }
-    data_type->flags = RUBY_TYPED_FREE_IMMEDIATELY;
-
-    c = TypedData_Make_Struct(rb_cData, RGObjClassInfo, data_type, cinfo);
+    data_type = rbgobj_class_info_create_data_type(klass);
+    c = TypedData_Make_Struct(rb_cObject, RGObjClassInfo, data_type, cinfo);
     cinfo->klass = klass;
     cinfo->gtype = gtype;
     cinfo->mark  = NULL;
@@ -229,9 +239,7 @@ get_superclass(GType gtype, VALUE module)
       case G_TYPE_OBJECT:
         return cInstantiatable;
       case G_TYPE_BOXED:
-        return rb_cObject;
       case G_TYPE_POINTER:
-        return rb_cData;
       case G_TYPE_ENUM:
       case G_TYPE_FLAGS:
         return rb_cObject;
@@ -368,7 +376,7 @@ rbgobj_class_info_lookup(VALUE klass)
     if (TYPE(klass) == T_CLASS) {
         VALUE super;
         super = rb_funcall(klass, id_superclass, 0);
-        return rbgobj_lookup_class(super);
+        return rbgobj_class_info_lookup(super);
     }
 
     rb_raise(rb_eRuntimeError, "can't get gobject class information");
@@ -420,6 +428,29 @@ rbgobj_gtype_to_ruby_class(GType gtype)
     return cinfo ? cinfo->klass : Qnil;
 }
 
+GType
+rbgobj_gtype_from_ruby(VALUE rb_gtype)
+{
+    if (NIL_P(rb_gtype)) {
+        return G_TYPE_NONE;
+    } else if (RB_TYPE_P(rb_gtype, RUBY_T_STRING)) {
+        GType gtype;
+        gtype = g_type_from_name(RVAL2CSTR(rb_gtype));
+        if (gtype == G_TYPE_INVALID) {
+            rb_raise(rb_eArgError,
+                     "unknown GType name: <%s>",
+                     RVAL2CSTR(rb_gtype));
+        }
+        return gtype;
+    } else if (RVAL2CBOOL(rb_obj_is_kind_of(rb_gtype, RG_TARGET_NAMESPACE))) {
+        return NUM2SIZET(rb_ivar_get(rb_gtype, id_gtype));
+    } else if (RVAL2CBOOL(rb_obj_is_kind_of(rb_gtype, rb_cClass))) {
+        return CLASS2GTYPE(rb_gtype);
+    } else {
+        return NUM2SIZET(rb_to_int(rb_gtype));
+    }
+}
+
 VALUE
 rbgobj_define_class(GType gtype, const gchar *name, VALUE module,
                     RGMarkFunc mark, RGFreeFunc free, VALUE parent)
@@ -431,7 +462,7 @@ rbgobj_define_class(GType gtype, const gchar *name, VALUE module,
 
     cinfo = (RGObjClassInfo *)rbgobj_class_info_lookup_by_gtype(gtype);
     if (cinfo) {
-        if (!rb_const_defined(module, rb_intern(name))) {
+        if (!rb_const_defined_at(module, rb_intern(name))) {
             rb_define_const(module, name, cinfo->klass);
             rbgobj_class_info_fill_name(cinfo);
         }
@@ -493,11 +524,13 @@ rbgobj_register_class(VALUE klass,
                       gboolean klass2gtype,
                       gboolean gtype2klass)
 {
-    RGObjClassInfo* cinfo = NULL;
+    rb_data_type_t *data_type = NULL;
+    RGObjClassInfo *cinfo = NULL;
     VALUE c = Qnil;
 
     if (klass2gtype) {
-        c = Data_Make_Struct(rb_cData, RGObjClassInfo, cinfo_mark, NULL, cinfo);
+        data_type = rbgobj_class_info_create_data_type(klass);
+        c = TypedData_Make_Struct(rb_cObject, RGObjClassInfo, data_type, cinfo);
     }
     if (gtype2klass && !cinfo)
         cinfo = g_new(RGObjClassInfo, 1);
@@ -508,6 +541,7 @@ rbgobj_register_class(VALUE klass,
         cinfo->mark  = NULL;
         cinfo->free  = NULL;
         cinfo->flags = 0;
+        cinfo->data_type = data_type;
     }
 
     if (klass2gtype)
@@ -571,26 +605,76 @@ init_typemap(void)
 /* GLib::Type */
 
 VALUE RG_TARGET_NAMESPACE;
-static ID id_gtype;
 
 VALUE
 rbgobj_gtype_new(GType gtype)
 {
     VALUE result = rb_obj_alloc(RG_TARGET_NAMESPACE);
-    VALUE arg = ULONG2NUM(gtype);
+    VALUE arg = SIZET2NUM(gtype);
     rb_obj_call_init(result, 1, &arg);
     return result;
 }
 
+/* Deprecated. Use rbgobj_gtype_from_ruby() instead. */
 GType
-rbgobj_gtype_get(VALUE self)
+rbgobj_gtype_get(VALUE rb_gtype)
 {
-    if (RVAL2CBOOL(rb_obj_is_kind_of(self, RG_TARGET_NAMESPACE))) {
-        return NUM2ULONG(rb_ivar_get(self, id_gtype));
-    } else {
-        return CLASS2GTYPE(self);
+    return rbgobj_gtype_from_ruby(rb_gtype);
+}
+
+static VALUE
+rg_s_try_convert(VALUE self, VALUE value)
+{
+    ID id_new;
+    CONST_ID(id_new, "new");
+
+    if (NIL_P(value))
+        return Qnil;
+
+    if (RVAL2CBOOL(rb_obj_is_kind_of(value, RG_TARGET_NAMESPACE)))
+        return value;
+
+    if (RVAL2CBOOL(rb_obj_is_kind_of(value, rb_cInteger))) {
+        GType gtype = NUM2ULONG(value);
+        if (!g_type_name(gtype))
+            return Qnil;
+        return rb_funcall(self, id_new, 1, value);
     }
-    rb_raise(rb_eTypeError, "Not a GLib::Type");
+
+    if (RB_TYPE_P(value, RUBY_T_SYMBOL)) {
+        value = rb_sym2str(value);
+    }
+
+    if (RB_TYPE_P(value, RUBY_T_STRING)) {
+        GType gtype = g_type_from_name(RVAL2CSTR(value));
+        if (gtype == G_TYPE_INVALID) {
+            return Qnil;
+        }
+        return rb_funcall(self, id_new, 1, value);
+    }
+
+    if (RVAL2CBOOL(rb_obj_is_kind_of(value, rb_cClass))) {
+        for (; !NIL_P(value); value = rb_funcall(value, id_superclass, 0)) {
+            RGObjClassInfo *cinfo;
+            VALUE data = rb_hash_aref(klass_to_cinfo, value);
+
+            if (NIL_P(data))
+                continue;
+
+            if (RTYPEDDATA_P(data)) {
+                TypedData_Get_Struct(data,
+                                     RGObjClassInfo,
+                                     RTYPEDDATA_TYPE(data),
+                                     cinfo);
+            } else {
+                Data_Get_Struct(data, RGObjClassInfo, cinfo);
+            }
+            return rb_funcall(self, id_new, 1, ULONG2NUM(cinfo->gtype));
+        }
+        return Qnil;
+    }
+
+    return Qnil;
 }
 
 static VALUE
@@ -599,7 +683,7 @@ rg_initialize(VALUE self, VALUE type)
     GType gtype;
 
     if (RVAL2CBOOL(rb_obj_is_kind_of(type, rb_cInteger))) {
-        gtype = NUM2ULONG(type);
+        gtype = NUM2SIZET(type);
         if (!g_type_name(gtype))
             gtype = G_TYPE_INVALID;
     } else {
@@ -609,7 +693,7 @@ rg_initialize(VALUE self, VALUE type)
     if (G_TYPE_INVALID == gtype)
         rb_raise(rb_eArgError, "invalid GType");
 
-    rb_ivar_set(self, id_gtype, ULONG2NUM(gtype));
+    rb_ivar_set(self, id_gtype, SIZET2NUM(gtype));
 
     return Qnil;
 }
@@ -617,7 +701,7 @@ rg_initialize(VALUE self, VALUE type)
 static VALUE
 rg_inspect(VALUE self)
 {
-    GType gtype = rbgobj_gtype_get(self);
+    GType gtype = rbgobj_gtype_from_ruby(self);
     gchar* str;
     VALUE result;
 
@@ -634,8 +718,8 @@ rg_operator_type_compare(VALUE self, VALUE other)
     if (!RVAL2CBOOL(rb_obj_is_kind_of(other, RG_TARGET_NAMESPACE)))
         return Qnil;
     else {
-        GType a = rbgobj_gtype_get(self);
-        GType b = rbgobj_gtype_get(other);
+        GType a = rbgobj_gtype_from_ruby(self);
+        GType b = rbgobj_gtype_from_ruby(other);
 
         if (a==b)
             return INT2FIX(0);
@@ -654,8 +738,8 @@ rg_operator_type_eq(VALUE self, VALUE other)
     if (!RVAL2CBOOL(rb_obj_is_kind_of(other, RG_TARGET_NAMESPACE)))
         return Qnil;
     else {
-        GType a = rbgobj_gtype_get(self);
-        GType b = rbgobj_gtype_get(other);
+        GType a = rbgobj_gtype_from_ruby(self);
+        GType b = rbgobj_gtype_from_ruby(other);
         return CBOOL2RVAL(a == b);
     }
 }
@@ -666,8 +750,8 @@ rg_operator_type_lt_eq(VALUE self, VALUE other)
     if (!RVAL2CBOOL(rb_obj_is_kind_of(other, RG_TARGET_NAMESPACE)))
         return Qnil;
     else {
-        GType a = rbgobj_gtype_get(self);
-        GType b = rbgobj_gtype_get(other);
+        GType a = rbgobj_gtype_from_ruby(self);
+        GType b = rbgobj_gtype_from_ruby(other);
         return CBOOL2RVAL(g_type_is_a(a, b));
     }
 }
@@ -678,8 +762,8 @@ rg_operator_type_gt_eq(VALUE self, VALUE other)
     if (!RVAL2CBOOL(rb_obj_is_kind_of(other, RG_TARGET_NAMESPACE)))
         return Qnil;
     else {
-        GType a = rbgobj_gtype_get(self);
-        GType b = rbgobj_gtype_get(other);
+        GType a = rbgobj_gtype_from_ruby(self);
+        GType b = rbgobj_gtype_from_ruby(other);
         return CBOOL2RVAL(g_type_is_a(b, a));
     }
 }
@@ -690,8 +774,8 @@ rg_operator_type_lt(VALUE self, VALUE other)
     if (!RVAL2CBOOL(rb_obj_is_kind_of(other, RG_TARGET_NAMESPACE)))
         return Qnil;
     else {
-        GType a = rbgobj_gtype_get(self);
-        GType b = rbgobj_gtype_get(other);
+        GType a = rbgobj_gtype_from_ruby(self);
+        GType b = rbgobj_gtype_from_ruby(other);
         return CBOOL2RVAL(g_type_is_a(a, b) && a != b);
     }
 }
@@ -702,8 +786,8 @@ rg_operator_type_gt(VALUE self, VALUE other)
     if (!RVAL2CBOOL(rb_obj_is_kind_of(other, RG_TARGET_NAMESPACE)))
         return Qnil;
     else {
-        GType a = rbgobj_gtype_get(self);
-        GType b = rbgobj_gtype_get(other);
+        GType a = rbgobj_gtype_from_ruby(self);
+        GType b = rbgobj_gtype_from_ruby(other);
         return CBOOL2RVAL(g_type_is_a(b, a) && a != b);
     }
 }
@@ -717,112 +801,113 @@ rg_hash(VALUE self)
 static VALUE
 rg_to_class(VALUE self)
 {
-    return GTYPE2CLASS(rbgobj_gtype_get(self));
+    return GTYPE2CLASS(rbgobj_gtype_from_ruby(self));
 }
 
 static VALUE
 rg_fundamental(VALUE self)
 {
-    return rbgobj_gtype_new(G_TYPE_FUNDAMENTAL(rbgobj_gtype_get(self)));
+    return rbgobj_gtype_new(G_TYPE_FUNDAMENTAL(rbgobj_gtype_from_ruby(self)));
 }
 
 static VALUE
 rg_fundamental_p(VALUE self)
 {
-    return CBOOL2RVAL(G_TYPE_IS_FUNDAMENTAL(rbgobj_gtype_get(self)));
+    return CBOOL2RVAL(G_TYPE_IS_FUNDAMENTAL(rbgobj_gtype_from_ruby(self)));
 }
 
 static VALUE
 rg_derived_p(VALUE self)
 {
-    return CBOOL2RVAL(G_TYPE_IS_DERIVED(rbgobj_gtype_get(self)));
+    return CBOOL2RVAL(G_TYPE_IS_DERIVED(rbgobj_gtype_from_ruby(self)));
 }
 
 static VALUE
 rg_interface_p(VALUE self)
 {
-    return CBOOL2RVAL(G_TYPE_IS_INTERFACE(rbgobj_gtype_get(self)));
+    return CBOOL2RVAL(G_TYPE_IS_INTERFACE(rbgobj_gtype_from_ruby(self)));
 }
 
 static VALUE
 rg_classed_p(VALUE self)
 {
-    return CBOOL2RVAL(G_TYPE_IS_CLASSED(rbgobj_gtype_get(self)));
+    return CBOOL2RVAL(G_TYPE_IS_CLASSED(rbgobj_gtype_from_ruby(self)));
 }
 
 static VALUE
 rg_instantiatable_p(VALUE self)
 {
-    return CBOOL2RVAL(G_TYPE_IS_INSTANTIATABLE(rbgobj_gtype_get(self)));
+    return CBOOL2RVAL(G_TYPE_IS_INSTANTIATABLE(rbgobj_gtype_from_ruby(self)));
 }
 
 static VALUE
 rg_derivable_p(VALUE self)
 {
-    return CBOOL2RVAL(G_TYPE_IS_DERIVABLE(rbgobj_gtype_get(self)));
+    return CBOOL2RVAL(G_TYPE_IS_DERIVABLE(rbgobj_gtype_from_ruby(self)));
 }
 
 static VALUE
 rg_deep_derivable_p(VALUE self)
 {
-    return CBOOL2RVAL(G_TYPE_IS_DEEP_DERIVABLE(rbgobj_gtype_get(self)));
+    return CBOOL2RVAL(G_TYPE_IS_DEEP_DERIVABLE(rbgobj_gtype_from_ruby(self)));
 }
 
 static VALUE
 rg_abstract_p(VALUE self)
 {
-    return CBOOL2RVAL(G_TYPE_IS_ABSTRACT(rbgobj_gtype_get(self)));
+    return CBOOL2RVAL(G_TYPE_IS_ABSTRACT(rbgobj_gtype_from_ruby(self)));
 }
 
 static VALUE
 rg_value_abstract_p(VALUE self)
 {
-    return CBOOL2RVAL(G_TYPE_IS_VALUE_ABSTRACT(rbgobj_gtype_get(self)));
+    return CBOOL2RVAL(G_TYPE_IS_VALUE_ABSTRACT(rbgobj_gtype_from_ruby(self)));
 }
 
 static VALUE
 rg_value_type_p(VALUE self)
 {
-    return CBOOL2RVAL(G_TYPE_IS_VALUE_TYPE(rbgobj_gtype_get(self)));
+    return CBOOL2RVAL(G_TYPE_IS_VALUE_TYPE(rbgobj_gtype_from_ruby(self)));
 }
 
 static VALUE
 rg_has_value_table(VALUE self)
 {
-    return CBOOL2RVAL(G_TYPE_HAS_VALUE_TABLE(rbgobj_gtype_get(self)));
+    return CBOOL2RVAL(G_TYPE_HAS_VALUE_TABLE(rbgobj_gtype_from_ruby(self)));
 }
 
 static VALUE
 rg_name(VALUE self)
 {
-    return rb_str_new2(g_type_name(rbgobj_gtype_get(self)));
+    return rb_str_new2(g_type_name(rbgobj_gtype_from_ruby(self)));
 }
 
 static VALUE
 rg_parent(VALUE self)
 {
-    GType parent = g_type_parent(rbgobj_gtype_get(self));
+    GType parent = g_type_parent(rbgobj_gtype_from_ruby(self));
     return parent ? rbgobj_gtype_new(parent) : Qnil;
 }
 
 static VALUE
 rg_depth(VALUE self)
 {
-    return UINT2NUM(g_type_depth(rbgobj_gtype_get(self)));
+    return UINT2NUM(g_type_depth(rbgobj_gtype_from_ruby(self)));
 }
 
 static VALUE
 rg_next_base(VALUE leaf_type, VALUE root_type)
 {
-    GType ret = g_type_next_base(rbgobj_gtype_get(leaf_type),
-                                 rbgobj_gtype_get(root_type));
+    GType ret = g_type_next_base(rbgobj_gtype_from_ruby(leaf_type),
+                                 rbgobj_gtype_from_ruby(root_type));
     return ret ? rbgobj_gtype_new(ret) : Qnil;
 }
 
 static VALUE
 rg_type_is_a_p(VALUE self, VALUE is_a_type)
 {
-    return CBOOL2RVAL(g_type_is_a(rbgobj_gtype_get(self), rbgobj_gtype_get(is_a_type)));
+    return CBOOL2RVAL(g_type_is_a(rbgobj_gtype_from_ruby(self),
+                                  rbgobj_gtype_from_ruby(is_a_type)));
 }
 
 #if 0
@@ -843,7 +928,7 @@ rg_children(VALUE self)
     VALUE result;
     guint i;
 
-    types = g_type_children(rbgobj_gtype_get(self), &n_children);
+    types = g_type_children(rbgobj_gtype_from_ruby(self), &n_children);
     result = rb_ary_new2(n_children);
     for (i = 0; i < n_children; i++)
         rb_ary_store(result, i, rbgobj_gtype_new(types[i]));
@@ -860,7 +945,7 @@ rg_interfaces(VALUE self)
     VALUE result;
     guint i;
 
-    types = g_type_interfaces(rbgobj_gtype_get(self), &n_interfaces);
+    types = g_type_interfaces(rbgobj_gtype_from_ruby(self), &n_interfaces);
     result = rb_ary_new2(n_interfaces);
     for (i = 0; i < n_interfaces; i++)
         rb_ary_store(result, i, rbgobj_gtype_new(types[i]));
@@ -873,7 +958,7 @@ static VALUE
 rg_class_size(VALUE self)
 {
     GTypeQuery query;
-    g_type_query(rbgobj_gtype_get(self), &query);
+    g_type_query(rbgobj_gtype_from_ruby(self), &query);
     return UINT2NUM(query.class_size);
 }
 
@@ -881,7 +966,7 @@ static VALUE
 rg_instance_size(VALUE self)
 {
     GTypeQuery query;
-    g_type_query(rbgobj_gtype_get(self), &query);
+    g_type_query(rbgobj_gtype_from_ruby(self), &query);
     return UINT2NUM(query.instance_size);
 }
 
@@ -889,6 +974,7 @@ static inline void
 _def_fundamental_type(VALUE ary, GType gtype, const char* name)
 {
     VALUE c = rbgobj_gtype_new(gtype);
+    rb_obj_freeze(c);
     rb_define_const(RG_TARGET_NAMESPACE, name, c);
     rb_ary_push(ary, c);
 }
@@ -898,10 +984,6 @@ _def_fundamental_type(VALUE ary, GType gtype, const char* name)
 void
 Init_gobject_gtype(void)
 {
-#if !GLIB_CHECK_VERSION(2, 35, 1)
-    g_type_init();
-#endif
-
     init_typemap();
 
     /* type */
@@ -918,6 +1000,8 @@ Init_gobject_gtype(void)
 
     RG_TARGET_NAMESPACE = rb_define_class_under(mGLib, "Type", rb_cObject);
 
+    RG_DEF_SMETHOD(try_convert, 1);
+
     rb_define_alias(CLASS_OF(RG_TARGET_NAMESPACE), "[]", "new");
     RG_DEF_METHOD(initialize, 1);
     RG_DEF_METHOD(inspect, 0);
